from typing import Callable, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from shipaw.fapi.requests import ShipmentRequest
    from shipaw.fapi.responses import ShipmentBookingResponse

CALLBACK_REGISTER: dict[str, Callable[['ShipmentRequest', 'ShipmentBookingResponse'], Any]] = {}


import os
from functools import wraps

from fastapi import APIRouter, Body
from fastapi.params import Depends
from starlette.requests import Request
from starlette.responses import HTMLResponse

from shipaw.config import ShipawSettings
from shipaw.fapi.alerts import Alerts
from shipaw.fapi.responses import ShipawTemplateResponse
from shipaw.fapi.form_data import shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import ShipmentRequest
from shipaw.models.shipment import Shipment, sample_shipment
from shipaw.fapi.routes_api import (
    order_results_api as order_confirm_json,
    order_summary_api as order_review_json,
    shipping_form_api as ship_form_json,
)

router = APIRouter()


def render_template_response(request: Request, resp: ShipawTemplateResponse) -> HTMLResponse:
    context = resp.template.context
    context['alerts'] = context.get('alerts', Alerts.empty()) + resp.alerts
    if resp.alerts.errors:
        return ShipawSettings.from_env().templates.TemplateResponse(
            request=request,
            name='alerts.html',
            context=context,
        )

    return ShipawSettings.from_env().templates.TemplateResponse(
        request=request,
        name=resp.template.template_path,
        context=context,
    )


def html_from_json(json_endpoint):
    @wraps(json_endpoint)
    async def wrapper(request: Request, *args, **kwargs):
        resp = await json_endpoint(request, *args, **kwargs)

        return ShipawSettings.from_env().templates.TemplateResponse(
            request=request, name=resp.template_path, context=resp.context
        )

    return wrapper


@router.post('/shipping_form', response_class=HTMLResponse)
async def shipping_form(request: Request, shipment: Shipment = Body(...)) -> HTMLResponse:
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.post('/order_summary', response_class=HTMLResponse)
async def order_summary(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> HTMLResponse:
    res = await order_review_json(request, shipment_request)
    return render_template_response(request, res)


@router.post('/order_results', response_class=HTMLResponse)
async def order_results(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> HTMLResponse:
    template_response = await order_confirm_json(request, shipment_request)
    return render_template_response(request, template_response)


@router.get('/test', response_class=HTMLResponse)
async def ship(request: Request) -> HTMLResponse:
    shipment = sample_shipment()
    res = await ship_form_json(request, shipment)
    return render_template_response(request, res)


@router.get('/home_mobile_phone', response_class=HTMLResponse)
async def home_mobile_phone():
    mobile_phone = ShipawSettings.from_env().mobile_phone
    return f"""
    <input type="tel" id="mobile_phone" name="mobile_phone" value="{mobile_phone}" required>
    """


@router.get('/open-file/{filepath}', response_class=HTMLResponse)
async def open_file(filepath: str):
    os.startfile(filepath)
    return HTMLResponse(content=f'<span>Re</span>')


@router.post('/print-file/{filepath}', response_class=HTMLResponse)
async def print_file(filepath: str):
    os.startfile(filepath, 'print')
    return HTMLResponse(content=f'<span>Re</span>')



from __future__ import annotations

from typing import ClassVar, override

from parcelforce_expresslink.address import (
    AddressRecipient,
    Contact as ContactPF,
)
from parcelforce_expresslink.combadge import CreateShipmentService
from parcelforce_expresslink.request_response import ShipmentRequest, ShipmentResponse as ShipmentResponsePF
#
from parcelforce_expresslink.config import ParcelforceSettings
from parcelforce_expresslink.client import ParcelforceClient
from parcelforce_expresslink.shipment import Shipment as ShipmentPF

from shipaw.models.logging import log_obj
from shipaw.providers.provider_abc import ShippingProvider
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.models.shipment import Shipment, Shipment as ShipmentAgnost
from shipaw.providers.parcelforce.provider_funcs import (
    PARCELFORCE_SERVICES,
    ParcelforceServices,
    address_from_agnostic_fc,
    contact_from_agnostic_fc,
    convert_shipment_by_direction,
    parcelforce_shipment_to_agnostic,
    ref_dict_from_str
)


# @dataclass
# @register_provider
class ParcelforceShippingProvider(ShippingProvider):
    name = 'PARCELFORCE'
    services: ClassVar[ParcelforceServices] = PARCELFORCE_SERVICES
    settings_type: ClassVar[type[ParcelforceSettings]] = ParcelforceSettings
    settings: ParcelforceSettings
    _client: ParcelforceClient | None = None

    def is_sandbox(self) -> bool:
        return 'test' in self.settings.pf_endpoint.lower()

    @property
    def client(self) -> ParcelforceClient:
        if self._client is None:
            if self.settings is None:
                raise ValueError('Settings must be set before using the client')
            self._client = ParcelforceClient(settings=self.settings)
        return self._client

    @override
    def provider_shipment(self, shipment: ShipmentAgnost) -> ShipmentPF:
        ship_pf = ShipmentPF(
            **ref_dict_from_str(shipment.reference),
            recipient_contact=contact_from_agnostic_fc(ContactPF, shipment.recipient),
            recipient_address=address_from_agnostic_fc(AddressRecipient, shipment.recipient),
            total_number_of_parcels=shipment.boxes,
            shipping_date=shipment.shipping_date,
            service_code=self.services.lookup(shipment.service),
            contract_number=self.settings.pf_contract_num_1,
            print_own_label=shipment.own_label,
        )
        convert_shipment_by_direction(ship_pf, shipment)
        return ship_pf

    @override
    def agnostic_shipment(self, shipment: ShipmentPF) -> Shipment:
        return parcelforce_shipment_to_agnostic(shipment)

    def build_booking_request(self, shipment: ShipmentAgnost) -> ShipmentRequest:
        shipment_pf = self.provider_shipment(shipment)
        shipment_request_pf = ShipmentRequest(requested_shipment=shipment_pf)
        log_obj(shipment_request_pf, 'ParcelForce Shipment Request')
        authorized_shipment = shipment_request_pf.authenticate(*self.settings.get_auth_secrets())
        return authorized_shipment

    @override
    def book_shipment(self, shipment: ShipmentAgnost) -> ShipmentBookingResponse:
        ship_req = self.build_booking_request(shipment)
        pf_response = self.make_pf_book_request(ship_req)
        return self.build_booking_response(pf_response, shipment)

    def build_booking_response(self, pf_response, shipment):
        """without label data"""
        return ShipmentBookingResponse(
            shipment=shipment,
            shipment_num=pf_response.shipment_num,
            tracking_link=self.settings.tracking_link(pf_response.shipment_num),
            data=pf_response.model_dump(),
            status=pf_response.status,
            success=pf_response.success,
        )

    def make_pf_book_request(self, ship_req):
        back = self.client.backend(CreateShipmentService)
        pf_response: ShipmentResponsePF = back.createshipment(request=ship_req)
        pf_response.handle_errors()
        return pf_response

    @override
    def fetch_label_content(self, shipment_num: str) -> bytes:
        return self.client.get_label_content(shipment_num)






from __future__ import annotations

from base64 import b64decode, b64encode
from pathlib import Path

from loguru import logger
from pawdf.array_pdf.array_p import on_a4
from pydantic import ConfigDict, Field, model_validator

from shipaw.config import ShipawSettings
from shipaw.models.base import ShipawBaseModel
from shipaw.fapi.alerts import Alerts
from shipaw.models.label_file import get_label_folder, get_label_stem, unused_path
from shipaw.models.shipment import Shipment


class ShipawTemplate(ShipawBaseModel):
    template_path: str
    context: dict = Field(default_factory=dict)

    def render_template(self, request):
        if not self.template_path:
            raise ValueError('No template_path set')
        return ShipawSettings.from_env().templates.TemplateResponse(
            request=request, name=self.template_path, context=self.context
        )


class BaseResponse(ShipawBaseModel):
    alerts: Alerts = Alerts.empty()
    data: dict | None = None
    success: bool | None = None
    status: str | None = None
    template: ShipawTemplate | None = None


class ShipmentBookingResponse(BaseResponse):
    shipment: Shipment
    shipment_num: str | None = None
    tracking_link: str | None = None
    label_data: bytes | None = None
    label_path: Path | None = None

    model_config = ConfigDict(json_encoders={bytes: lambda v: b64encode(v).decode('utf-8') if v else None})


    @model_validator(mode='after')
    def get_label_path(self):
        if self.label_path is None:
            folder = get_label_folder(self.shipment.direction)
            label_stem = get_label_stem(self.shipment)
            label_filepath = (folder / label_stem).with_suffix('.pdf')
            self.label_path = unused_path(label_filepath)
        return self

    async def write_label_file(self):
        try:
            label_content = self.label_data
        except Exception as e:
            logger.error(f'Error getting label content: {e}')
            raise
        label_path = self.label_path
        unsize = label_path.parent / 'original_size' / label_path.name
        unsize.parent.mkdir(parents=True, exist_ok=True)
        unsize.write_bytes(label_content)
        on_a4(input_file=unsize, output_file=label_path)
        logger.info(f'Wrote label to {label_path}')


class ShipawTemplateResponse(BaseResponse):
    template: ShipawTemplate



from pydantic import BaseModel

from royal_mail_click_and_drop.models.address import AddressRequest as AddressRM, RecipientDetailsRequest as ContactRM
from royal_mail_click_and_drop.models.create_orders_request import CreateOrderRequest
from shipaw.models.address import Address, Contact, FullContact
from shipaw.models.services import Services
from shipaw.models.shipment import Shipment
from shipaw.providers.provider_abc import ShippingProvider

# ServiceDictPlaceholder: ServiceDict = {
#     'NEXT_DAY': 'sdfdasfg',
#     'NEXT_DAY_12': 'sgsdg',
#     'NEXT_DAY_9': 'sdgsd',
# }

SERVICES_PLACEHOLDER = Services(
    NEXT_DAY='',
    NEXT_DAY_12='',
    NEXT_DAY_9='',
)


def rm_address_from_agnostic_fc(full_contact: FullContact) -> AddressRM:
    return AddressRM(
        full_name=full_contact.contact.contact_name,
        company_name=full_contact.address.business_name,
        address_line1=full_contact.address.address_lines[0],
        address_line2=full_contact.address.address_lines[1] if len(full_contact.address.address_lines) > 1 else None,
        address_line3=full_contact.address.address_lines[2] if len(full_contact.address.address_lines) > 2 else None,
        city=full_contact.address.town,
        county=full_contact.address.county,
        postcode=full_contact.address.postcode,
        country_code=full_contact.address.country,
    )


def rm_contact_from_agnostic_fc(full_contact: FullContact) -> ContactRM:
    return ContactRM(
        address=rm_address_from_agnostic_fc(full_contact),
        phone_number=full_contact.contact.mobile_phone or full_contact.contact.phone_number,
        email_address=full_contact.contact.email_address,
    )


def full_contact_from_rm(recipient: ContactRM) -> FullContact:
    return FullContact(
        contact=Contact(
            contact_name=recipient.address.full_name,
            phone_number=recipient.phone_number,
            email_address=recipient.email_address,
            mobile_phone=recipient.phone_number,
        ),
        address=Address(
            business_name=recipient.address.company_name,
            address_lines=[
                line
                for line in [
                    recipient.address.address_line1,
                    recipient.address.address_line2,
                    recipient.address.address_line3,
                ]
                if line
            ],
            town=recipient.address.city,
            county=recipient.address.county,
            postcode=recipient.address.postcode,
            country=recipient.address.country_code,
        ),
    )


def shipment_from_rm(shipment: CreateOrderRequest) -> Shipment:
    return Shipment()


class RoyalMailProvider(ShippingProvider):
    name: str = 'ROYAL_MAIL'
    services = SERVICES_PLACEHOLDER

    def provider_shipment(self, shipment: Shipment) -> BaseModel:
        pass

    def agnostic_shipment(self, shipment: BaseModel) -> Shipment:
        pass

    def book_shipment(self, shipment: dict | Shipment) -> 'ShipmentBookingResponseAgnost': ...

    def fetch_label_content(self, shipment_num: str) -> bytes: ...


from pathlib import Path
from typing import cast

from combadge.core.errors import BackendError
from fastapi import APIRouter, Body, Depends
from loguru import logger
from starlette.requests import Request
from starlette.responses import JSONResponse
from starlette.staticfiles import StaticFiles

from shipaw.config import ShipawSettings
from shipaw.fapi.alerts import Alert, AlertType, Alerts, maybe_alert_phone_number
from shipaw.fapi.backend import try_book_shipment, try_get_write_label, maybe_alert_apc, convert_choice
from shipaw.fapi.form_data import shipment_request_form, shipment_request_form_json
from shipaw.fapi.requests import AddressRequest, ShipmentRequest
from shipaw.fapi.responses import ShipawTemplate, ShipawTemplateResponse
from shipaw.models.address import Address, AddressChoice as AddressChoiceAgnost
from shipaw.models.logging import log_obj
from shipaw.models.shipment import Shipment
from shipaw.providers.parcelforce.provider import ParcelforceShippingProvider
from shipaw.providers.parcelforce.provider_funcs import (
    address_from_agnostic,
)
from shipaw.providers.registry import PROVIDER_REGISTER

router = APIRouter()
router.mount('/static', StaticFiles(directory=str(ShipawSettings.from_env().static_dir)), name='static')


@router.post('/shipping_form', response_model=ShipawTemplateResponse)
async def shipping_form_api(request: Request, shipment: Shipment = Body(...)) -> ShipawTemplateResponse:
    log_obj(shipment, 'Shipment received at /ship_form:')
    alerts: Alerts = request.app.alerts

    if any(['prdev' in str(_).lower() for _ in Path(__file__).parents]):
        msg = '"prdev" in cwd tree - BETA MODE - This is a development version'
        logger.warning(msg)
        alerts += Alert(message=msg, type=AlertType.WARNING)

    if ShipawSettings.from_env().shipper_live:
        msg = 'Shipper Live is True - Real Shipments will be booked'
    else:
        msg = 'Shipper_live is False - No Shipments will be booked'
    logger.warning(msg)
    alerts += Alert(message=msg, type=AlertType.NOTIFICATION)

    tmplt = ShipawTemplate(template_path='shipping_form_container.html', context={'shipment': shipment})
    return ShipawTemplateResponse(template=tmplt, alerts=alerts)


@router.post('/order_summary', response_model=ShipawTemplateResponse)
async def order_summary_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form),
) -> ShipawTemplateResponse:
    log_obj(shipment_request, 'ShipmentRequest received at shipaw/order_summary:')
    context = {'shipment_request': shipment_request}

    alerts = await maybe_alert_phone_number(shipment_request.shipment.remote_full_contact.contact.mobile_phone)
    alerts += await maybe_alert_apc(shipment_request)

    return ShipawTemplateResponse(
        template=ShipawTemplate(template_path='/order_summary.html', context=context),
        alerts=alerts,
    )


@router.post('/order_results', response_model=ShipawTemplateResponse)
async def order_results_api(
    request: Request,
    shipment_request: ShipmentRequest = Depends(shipment_request_form_json),
) -> ShipawTemplateResponse:
    shipment_response = await try_book_shipment(shipment_request)
    await try_get_write_label(shipment_request, shipment_response)

    if shipment_response.alerts.errors:
        log_obj(shipment_response.alerts, 'Errors booking shipment:')
        alerts = shipment_response.alerts
        shipment_response.template = ShipawTemplate(
            template_path='/alerts.html',
            context={'alerts': alerts},
        )
        return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)

    log_obj(shipment_response, 'Shipment Booked')
    await try_get_write_label(shipment_request, shipment_response)

    if hasattr(request.app, 'callback'):
        await request.app.callback(shipment_request, shipment_response)

    shipment_response.template = ShipawTemplate(
        template_path='/order_results.html',
        context={'response': shipment_response},
    )
    return ShipawTemplateResponse.model_validate(shipment_response, from_attributes=True)


@router.post('/addr_choices', response_model=list[AddressChoiceAgnost], response_class=JSONResponse)
async def get_addr_choices_api(
    request: Request,
    body: AddressRequest = Body(...),
) -> list[AddressChoiceAgnost]:
    """Fetch candidate address choices for a postcode, optionally scored by closeness to provided address.
    Hardcoded to use Parcelforce provider for now - APC does not provide address lookup.

    Args:
        request: Request - FastAPI request object
        body: Address - request body containing postcode and optional address
    """
    p: ParcelforceShippingProvider = cast(ParcelforceShippingProvider, PROVIDER_REGISTER['PARCELFORCE'])
    client = p.client
    postcode = body.postcode
    address_agnost = body.address
    pf_address = address_from_agnostic(address_agnost) if address_agnost else None
    # log_obj(pf_address, 'Address received at /cand:')

    try:
        res_pf = client.get_choices(postcode=postcode, address=pf_address)
        res = [await convert_choice(_) for _ in res_pf]
        return res

    except BackendError as e:
        alert = Alert(
            message=f'Error fetching candidates: {e}',
            type=AlertType.ERROR,
        )
        request.app.alerts += alert
        logger.warning(f'Error fetching candidates: {e}')
        addr = Address(address_lines=['ERROR:', str(e)], town='Error', postcode='Error', business_name='Error')
        chc = AddressChoiceAgnost(address=addr, score=0)
        return [chc]





# from shipaw.providers.apc.provider import APCShippingProvider
# from shipaw.providers.parcelforce.provider import ParcelforceShippingProvider
#
# __all__ = ['APCShippingProvider', 'ParcelforceShippingProvider']


import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import ClassVar, Self, TYPE_CHECKING

from loguru import logger
from pydantic import BaseModel
from pydantic_settings import BaseSettings

from shipaw.config import ShipawSettings
from shipaw.models.base import ShipawBaseModel
from shipaw.models.logging import log_obj
from shipaw.models.services import Services
from shipaw.models.shipment import Shipment

if TYPE_CHECKING:
    from shipaw.fapi.requests import ShipmentRequest
    from shipaw.fapi.responses import ShipmentBookingResponse


class ShippingProvider(ABC, ShipawBaseModel):
    name: ClassVar[str]
    services: ClassVar[Services]
    settings_type: ClassVar[type[BaseSettings]]
    settings: BaseSettings

    @abstractmethod
    def is_sandbox(self) -> bool: ...

    @classmethod
    def from_env(cls, env_file: Path) -> Self:
        settings = cls.settings_type(_env_file=env_file)
        return cls(settings=settings)

    @classmethod
    def from_env_settings(cls, shipaw_settings=None) -> Self:
        shipaw_settings = shipaw_settings or ShipawSettings.from_env('SHIPAW_ENV')
        env_file = shipaw_settings.provider_env_dict.get(cls.name)
        provider_settings = cls.settings_type(_env_file=env_file)
        return cls(settings=provider_settings)

    @staticmethod
    @abstractmethod
    def provider_shipment(shipment: Shipment) -> BaseModel:
        """Takes agnostic Shipment object and returns provider Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def agnostic_shipment(shipment: BaseModel) -> Shipment:
        """Takes provider Shipment object and returns agnostic Shipment object"""
        ...

    @staticmethod
    @abstractmethod
    def book_shipment(shipment: dict | Shipment) -> 'ShipmentBookingResponse': ...

    @staticmethod
    @abstractmethod
    def fetch_label_content(shipment_num: str) -> bytes: ...

    # @staticmethod
    # async def booking_response_callback(request: 'ShipmentRequest', response: 'ShipmentBookingResponse'):
    #     """Do after booking, e.g. log, write label file, etc."""
    #     log_obj(response, 'Shipment Booked')
    #     try:
    #         if response.label_data is None and response.shipment_num:
    #             logger.info('Fetching missing label data...')
    #             response.label_data = request.provider.fetch_label_content(response.shipment_num)
    #         await response.write_label_file()
    #     except Exception as e:
    #         logger.exception(f'Error getting or writing label data: {e}')

    def wait_fetch_label(self, shipment_num: str) -> bytes:
        for i in range(10):
            try:
                time.sleep(1)
                label_data = self.fetch_label_content(shipment_num=shipment_num)
                assert label_data is not None
                return label_data
            except AssertionError as e:
                print(f'Label not ready yet for {shipment_num}, retrying...')
        raise RuntimeError(f'Label not ready after retries for {shipment_num}')


    async def wait_fetch_label_as(self, shipment_num: str) -> bytes:
        for i in range(10):
            try:
                time.sleep(1)
                label_data = self.fetch_label_content(shipment_num=shipment_num)
                assert label_data is not None
                return label_data
            except AssertionError as e:
                print(f'Label not ready yet for {shipment_num}, retrying...')
        raise RuntimeError(f'Label not ready after retries for {shipment_num}')




from apc_hypaship.error import apc_http_status_alerts
from httpx import HTTPStatusError
from loguru import logger
from parcelforce_expresslink.address import AddressChoice as AddressChoicePF, Contact as ContactPF

from shipaw.fapi.requests import ShipmentRequest
from shipaw.fapi.responses import ShipmentBookingResponse
from shipaw.fapi.alerts import Alert, Alerts, AlertType
from shipaw.models.address import AddressChoice as AddressChoiceAgnost
from shipaw.models.ship_types import ShipDirection
from shipaw.providers.parcelforce.provider_funcs import full_contact_from_provider_contact_address


async def try_book_shipment(shipment_request: ShipmentRequest) -> ShipmentBookingResponse:
    shipment_response = ShipmentBookingResponse(shipment=shipment_request.shipment)
    try:
        shipment_response = shipment_request.provider.book_shipment(shipment_request.shipment)

    except HTTPStatusError as e:
        await maybe_apc(e, shipment_request, shipment_response)

    except Exception as e:
        logger.exception(f'Error booking shipment: {e}')
        shipment_response.alerts += Alert.from_exception(e)

    return shipment_response


async def try_get_write_label(request: ShipmentRequest, response: ShipmentBookingResponse):
    if not response.label_data:
        await try_get_label_data(request, response)
    await try_write_label(response)


async def try_get_label_data(request: ShipmentRequest, response: ShipmentBookingResponse) -> None:
    try:
        if response.label_data is not None:
            logger.info('Label data already present, not fetching')
        else:
            if response.shipment_num:
                response.label_data = await request.provider.wait_fetch_label_as(response.shipment_num)
            else:
                logger.warning('No shipment number to fetch label data')

    except HTTPStatusError as e:
        await maybe_apc(e, request, response)

    except Exception as e:
        logger.exception(f'Error getting label data')
        response.alerts += Alert.from_exception(e)


async def try_write_label(response: ShipmentBookingResponse):
    try:
        await response.write_label_file()
    except Exception as e:
        logger.exception(f'Error writing label file: {e}')
        response.alerts += Alert.from_exception(e)


async def maybe_apc(e: HTTPStatusError, shipment_request, shipment_response):
    if 'APC' in shipment_request.provider_name:
        for alert in await apc_http_status_alerts(e):
            shipment_response.alerts += alert
    else:
        logger.exception(e)
        shipment_response.alerts += Alert.from_exception(e)


async def maybe_alert_apc(shipment_request):
    alerts = Alerts.empty()
    if shipment_request.provider_name == 'APC' and shipment_request.shipment.direction == ShipDirection.DROPOFF:
        alerts += Alert(
            message='APC does not support drop-off shipments - please select Outbound or Inbound Collection',
            type=AlertType.ERROR,
        )
    return alerts


async def convert_choice(choice: AddressChoicePF) -> AddressChoiceAgnost:
    fc = full_contact_from_provider_contact_address(contact=ContactPF.empty(), address=choice.address)
    return AddressChoiceAgnost(address=fc.address, score=choice.score)
